=== modified file 'bzrlib/remote.py'
--- bzrlib/remote.py	2008-06-07 22:15:25 +0000
+++ bzrlib/remote.py	2008-06-17 04:32:18 +0000
@@ -840,7 +840,7 @@
     def _get_parent_map(self, keys):
         """Helper for get_parent_map that performs the RPC."""
         medium = self._client._medium
-        if not medium._remote_is_at_least_1_2:
+        if medium._is_remote_before((1, 2)):
             # We already found out that the server can't understand
             # Repository.get_parent_map requests, so just fetch the whole
             # graph.
@@ -919,7 +919,7 @@
             medium.disconnect()
             # To avoid having to disconnect repeatedly, we keep track of the
             # fact the server doesn't understand remote methods added in 1.2.
-            medium._remote_is_at_least_1_2 = False
+            medium._remember_remote_is_before((1, 2))
             return self.get_revision_graph(None)
         response_tuple, response_handler = response
         if response_tuple[0] not in ['ok']:
@@ -1080,7 +1080,7 @@
 
     def get_data_stream_for_search(self, search):
         medium = self._client._medium
-        if not medium._remote_is_at_least_1_2:
+        if medium._is_remote_before((1, 2)):
             self._ensure_real()
             return self._real_repository.get_data_stream_for_search(search)
         REQUEST_NAME = 'Repository.stream_revisions_chunked'
@@ -1101,7 +1101,7 @@
             medium.disconnect()
             # To avoid having to disconnect repeatedly, we keep track of the
             # fact the server doesn't understand this remote method.
-            medium._remote_is_at_least_1_2 = False
+            medium._remember_remote_is_before((1, 2))
             self._ensure_real()
             return self._real_repository.get_data_stream_for_search(search)
 

=== modified file 'bzrlib/smart/client.py'
--- bzrlib/smart/client.py	2008-05-30 04:33:39 +0000
+++ bzrlib/smart/client.py	2008-06-17 04:22:26 +0000
@@ -61,6 +61,9 @@
                     response_handler)
         else:
             for protocol_version in [3, 2]:
+                if protocol_version == 2:
+                    # If v3 doesn't work, the remote side is older than 1.6.
+                    self._medium._remember_remote_is_before((1, 6))
                 response_handler = self._send_request(
                     protocol_version, method, args, body=body,
                     readv_body=readv_body)

=== modified file 'bzrlib/smart/medium.py'
--- bzrlib/smart/medium.py	2008-05-20 23:30:28 +0000
+++ bzrlib/smart/medium.py	2008-06-17 04:32:18 +0000
@@ -440,9 +440,45 @@
         self._protocol_version = None
         self._done_hello = False
         # Be optimistic: we assume the remote end can accept new remote
-        # requests until we get an error saying otherwise.  (1.2 adds some
-        # requests that send bodies, which confuses older servers.)
-        self._remote_is_at_least_1_2 = True
+        # requests until we get an error saying otherwise.
+        # _remote_version_is_before tracks the bzr version the remote side
+        # can be based on what we've seen so far.
+        self._remote_version_is_before = None
+
+    def _is_remote_before(self, version_tuple):
+        """Is it possible the remote side is supports RPCs for a given version?
+
+        Typical use::
+
+            needed_version = (1, 2)
+            if medium._is_remote_before(needed_version):
+                fallback_to_pre_1_2_rpc()
+            else:
+                try:
+                    do_1_2_rpc()
+                except UnknownSmartMethod:
+                    medium._remember_remote_is_before(needed_version)
+                    fallback_to_pre_1_2_rpc()
+
+        :seealso: _remember_remote_is_before
+        """
+        if self._remote_version_is_before is None:
+            # So far, the remote side seems to support everything
+            return False
+        return version_tuple >= self._remote_version_is_before
+
+    def _remember_remote_is_before(self, version_tuple):
+        """Tell this medium that the remote side is older the given version.
+
+        :seealso: _is_remote_before
+        """
+        if (self._remote_version_is_before is not None and
+            version_tuple > self._remote_version_is_before):
+            raise AssertionError, (
+                "_remember_remote_is_before(%r) called, but "
+                "_remember_remote_is_before(%r) was called previously."
+                % (version_tuple, self._remote_version_is_before))
+        self._remote_version_is_before = version_tuple
 
     def protocol_version(self):
         """Find out if 'hello' smart request works."""

=== modified file 'bzrlib/tests/test_remote.py'
--- bzrlib/tests/test_remote.py	2008-05-22 05:48:22 +0000
+++ bzrlib/tests/test_remote.py	2008-06-17 04:32:18 +0000
@@ -187,9 +187,8 @@
 class FakeMedium(medium.SmartClientMedium):
 
     def __init__(self, client_calls, base):
-        self._remote_is_at_least_1_2 = True
+        medium.SmartClientMedium.__init__(self, base)
         self._client_calls = client_calls
-        self.base = base
 
     def disconnect(self):
         self._client_calls.append(('disconnect medium',))
@@ -254,6 +253,34 @@
                 'xyz/', scheme + '//host/path', 'xyz/')
 
 
+class Test_ClientMedium_remote_is_at_least(tests.TestCase):
+    """Tests for the behaviour of client_medium.remote_is_at_least."""
+
+    def test_initially_unlimited(self):
+        """A fresh medium assumes that the remote side supports all
+        versions.
+        """
+        client_medium = medium.SmartClientMedium('dummy base')
+        self.assertFalse(client_medium._is_remote_before((99, 99)))
+    
+    def test__remember_remote_is_before(self):
+        """Calling _remember_remote_is_before ratchets down the known remote
+        version.
+        """
+        client_medium = medium.SmartClientMedium('dummy base')
+        # Mark the remote side as being less than 1.6.  The remote side may
+        # still be 1.5.
+        client_medium._remember_remote_is_before((1, 6))
+        self.assertTrue(client_medium._is_remote_before((1, 6)))
+        self.assertFalse(client_medium._is_remote_before((1, 5)))
+        # Calling _remember_remote_is_before again with a lower value works.
+        client_medium._remember_remote_is_before((1, 5))
+        self.assertTrue(client_medium._is_remote_before((1, 5)))
+        # You cannot call _remember_remote_is_before with a larger value.
+        self.assertRaises(
+            AssertionError, client_medium._remember_remote_is_before, (1, 9))
+
+
 class TestBzrDirOpenBranch(tests.TestCase):
 
     def test_branch_present(self):
@@ -883,7 +910,7 @@
         repo, client = self.setup_fake_client_and_repository(transport_path)
         client.add_unknown_method_response('Repository,get_parent_map')
         client.add_success_response_with_body('', 'ok')
-        self.assertTrue(client._medium._remote_is_at_least_1_2)
+        self.assertFalse(client._medium._is_remote_before((1, 2)))
         rev_id = 'revision-id'
         expected_deprecations = [
             'bzrlib.remote.RemoteRepository.get_revision_graph was deprecated '
@@ -898,7 +925,7 @@
               ('quack/', ''))],
             client._calls)
         # The medium is now marked as being connected to an older server
-        self.assertFalse(client._medium._remote_is_at_least_1_2)
+        self.assertTrue(client._medium._is_remote_before((1, 2)))
 
     def test_get_parent_map_fallback_parentless_node(self):
         """get_parent_map falls back to get_revision_graph on old servers.  The
@@ -915,7 +942,7 @@
         transport_path = 'quack'
         repo, client = self.setup_fake_client_and_repository(transport_path)
         client.add_success_response_with_body(rev_id, 'ok')
-        client._medium._remote_is_at_least_1_2 = False
+        client._medium._remember_remote_is_before((1, 2))
         expected_deprecations = [
             'bzrlib.remote.RemoteRepository.get_revision_graph was deprecated '
             'in version 1.4.']

=== modified file 'bzrlib/tests/test_smart_transport.py'
--- bzrlib/tests/test_smart_transport.py	2008-06-02 01:12:17 +0000
+++ bzrlib/tests/test_smart_transport.py	2008-06-17 04:32:18 +0000
@@ -2774,7 +2774,7 @@
         self.assertCallDoesNotBreakMedium('method', ('args',), u'body')
 
 
-class MockMedium(object):
+class MockMedium(medium.SmartClientMedium):
     """A mock medium that can be used to test _SmartClient.
     
     It can be given a series of requests to expect (and responses it should
@@ -2791,10 +2791,9 @@
     """
 
     def __init__(self):
-        self.base = 'dummy base'
+        super(MockMedium, self).__init__('dummy base')
         self._mock_request = _MockMediumRequest(self)
         self._expected_events = []
-        self._protocol_version = None
         
     def expect_request(self, request_bytes, response_bytes,
                        allow_partial_read=False):
@@ -2931,6 +2930,9 @@
         # medium, and the smart_client returns the response from the server.
         self.assertEqual(('response value',), result)
         self.assertEqual([], medium._expected_events)
+        # Also, the v3 works then the server should be assumed to support RPCs
+        # introduced in 1.6.
+        self.assertFalse(medium._is_remote_before((1, 6)))
 
     def test_version_two_server(self):
         """If the server only speaks protocol 2, the client will first try
@@ -2969,6 +2971,10 @@
         self.assertEqual(('another response',), result)
         self.assertEqual([], medium._expected_events)
 
+        # Also, because v3 is not supported, the client medium should assume
+        # that RPCs introduced in 1.6 aren't supported either.
+        self.assertTrue(medium._is_remote_before((1, 6)))
+
     def test_unknown_version(self):
         """If the server does not use any known (or at least supported)
         protocol version, a SmartProtocolError is raised.

=== modified file 'bzrlib/transport/http/_urllib.py'
--- bzrlib/transport/http/_urllib.py	2008-05-19 10:04:15 +0000
+++ bzrlib/transport/http/_urllib.py	2008-05-29 06:43:43 +0000
@@ -48,7 +48,6 @@
             self._opener = _from_transport._opener
         else:
             self._opener = self._opener_class()
-        self._remote_is_at_least_1_2 = True
 
     def _perform(self, request):
         """Send the request to the server and handles common errors.

